"""
Created on 2024-10-08

@author: wf
"""

import os
import traceback

from nomina.file_formats import AccountingFileFormats
from tests.basetest import Basetest
from nomina.nomina_cmd import NominaCmd

class TestNominaConverter(Basetest):
    """
    test the nomina command line converter
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.input_files = [
            "empty.yaml",
            "empty_xml.gnucash",
            "example.beancount",
            "expenses.qif",
            "expenses.yaml",
            "expenses2024_bzv.yaml",
            "expenses2024.yaml",
            "expenses2024_xml.gnucash",
            "expenses_xml.gnucash",
            "simple_sample.yaml",
            "simple_sample_xml.gnucash",
        ]
        self.target_formats = [
            ("LB-YAML", ".yaml"),
            ("GC-XML", ".gnucash"),
            ("BEAN", ".beancount"),
        ]
        self.formats = AccountingFileFormats()
        self.target_dir = "/tmp/nomina"
        os.makedirs(self.target_dir, exist_ok=True)

        self.cmd = NominaCmd()

    def test_conversions(self):
        """
        test the conversions
        """
        for input_file in self.input_files:
            for target_format, ext in self.target_formats:
                with self.subTest(
                    input_file=input_file, target_format=target_format, ext=ext
                ):
                    input_path = self.examples_path + f"/{input_file}"
                    from_format=self.formats.detect_format(input_path)
                    if from_format.acronym!=target_format:
                        self.run_conversion(
                            input_path, target_format=target_format, ext=ext
                        )

    def run_conversion(self, input_path, target_format: str, ext: str):
        """
        run the given conversion via command line
        """
        base_name = os.path.basename(input_path)
        name, _ = os.path.splitext(base_name)
        converted_name = f"{name}_converted{ext}"
        output_file = os.path.join(self.target_dir, converted_name)

        argv = [
            "--debug",
            "--convert", input_path,
            "--format", target_format,
            "--output", output_file
        ]

        try:
            exit_code = self.cmd.cmd_main(argv)
            self.assertEqual(exit_code, 0, f"Command failed with exit code {exit_code}")
            self.assertTrue(os.path.exists(output_file), f"Output file not created: {output_file}")
            file_size=os.path.getsize(output_file)
            self.assertGreater(file_size, 0, f"Output file has size {file_size}: {output_file}")
        except Exception as e:
            if self.debug:
                print(traceback.format_exc())
            self.fail(f"Conversion failed for {input_path} to {target_format}:\n{str(e)}")


"""
Created on 2024-10-06

@author: wf
"""
from pathlib import Path
from typing import Type, Dict
from nomina.file_formats import AccountingFileFormats
from nomina.ledger import Book
from nomina.beancount_ledger import BeancountToLedgerConverter, LedgerToBeancountConverter
from nomina.bzv_ledger import BankingZVToLedgerConverter
from nomina.gnc_ledger import GnuCashToLedgerConverter, LedgerToGnuCashConverter
from nomina.qif_ledger import QifToLedgerConverter
from nomina.nomina_converter import BaseToLedgerConverter, BaseFromLedgerConverter

class Converter:
    """
    General converter for personal accounting formats using hub and spoke model
    """

    def __init__(self, args):
        self.args = args
        self.detector = AccountingFileFormats()
        self.to_ledger: Dict[str, Type[BaseToLedgerConverter]] = {
            "GC-XML": GnuCashToLedgerConverter,
            "QIF": QifToLedgerConverter,
            "BEAN": BeancountToLedgerConverter,
            "BZV-YAML": BankingZVToLedgerConverter,
            "LB-YAML": None,
        }
        self.from_ledger: Dict[str, Type[BaseFromLedgerConverter]] = {
            "GC-XML": LedgerToGnuCashConverter,
            "BEAN": LedgerToBeancountConverter,
            "LB-YAML":  None,
        }

    def convert(self, input_path: Path = None, output_format: str = None) -> None:
        """
        Convert the input file to the specified output format.

        Args:
            input_path (str, optional): Path to the input file. Defaults to None.
            output_format (str, optional): Acronym of the output format. Defaults to None.
        """
        if input_path is None:
            input_path = self.args.convert
        if output_format is None:
            output_format = self.args.format

        input_format = self.detector.detect_format(input_path)

        if not input_format:
            raise ValueError(f"Unsupported or unrecognized input format for file: {input_path}")

        if input_format.acronym not in self.to_ledger:
            raise ValueError(f"Unsupported input format: {input_format.acronym}")
        if output_format not in self.from_ledger:
            raise ValueError(f"Unsupported output format: {output_format}")

        to_ledger_cls = self.to_ledger[input_format.acronym]
        from_ledger_cls = self.from_ledger[output_format]

        # Convert to LedgerBook
        if to_ledger_cls is not None:
            to_ledger = to_ledger_cls(debug=self.args.debug)
            ledger_book = to_ledger.convert_to_ledger(input_path)
        else:
            # If input is already LedgerBook, just load it
            ledger_book = Book.load_from_yaml_file(input_path)

        # Convert from LedgerBook to output format
        if from_ledger_cls is not None:
            from_ledger = from_ledger_cls(debug=self.args.debug)
            _target_object = from_ledger.convert_from_ledger(ledger_book)
            output_text = from_ledger.to_text()
            pass
        else:
            # target is a LedgerBook get the XML markup
            output_text = ledger_book.to_yaml()
        output_stream = self.args.output.open('w')
        output_stream.write(output_text)
        output_stream.close()


    def get_supported_formats(self) -> Dict[str, list]:
        """
        Get a list of supported input and output formats

        Returns:
            Dict[str, list]: A dictionary with 'input' and 'output' keys, each containing a list of supported format acronyms
        """
        input_formats = set(self.to_ledger.keys())
        output_formats = set(self.from_ledger.keys())
        return {"input": list(input_formats), "output": list(output_formats)}



"""
Created on 2024-10-08

@author: wf
"""

from pathlib import Path

from nomina.bzv import Book
from nomina.bzv_ledger import BankingZVToLedgerConverter
from nomina.stats import Stats
from tests.basetest import Basetest
from tests.example_testcases import NominaExample


class Test_Ledger_BankingZV(Basetest):
    """
    test Banking ZV Subsembly handling
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_banking_zv_to_ledger(self):
        """
        test converting Banking ZV format to ledger
        """
        bzv_example = self.examples_path + "/expenses2024_bzv.yaml"
        converter = BankingZVToLedgerConverter(debug=self.debug)
        banking_zv_book = converter.load(bzv_example)
        self.assertIsInstance(banking_zv_book, Book)
        ledger_book = converter.convert_to_target()
        stats = ledger_book.get_stats()
        if self.debug:
            stats.show()
        example = NominaExample(
            name="expenses_2024",
            owner="John Doe",
            url=None,
            example_path=Path(self.examples_path),
            do_log=True,
            expected_stats=Stats(
                accounts=3,
                transactions=2,
                start_date="2024-10-06",
                end_date="2024-10-06",
                currencies={
                    "EUR": 3,
                },
            ),
        )
        wrong = example.check_stats(stats)
        self.assertEqual(0, wrong)



"""
Created on 2024-10-07

@author: wf
"""

import os
from pathlib import Path

from nomina.nomina_beancount import Beancount
from nomina.beancount_ledger import (
    BeancountToLedgerConverter,
    LedgerToBeancountConverter,
)
from nomina.stats import Stats
from tests.basetest import Basetest
from tests.example_testcases import NominaExample


class Test_LedgerBeancount(Basetest):
    """
    test Beancount handling
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.examples = NominaExample.get_examples()

    def test_beancount2ledger(self):
        """
        test converting beancount format to ledger
        """
        bc_example = self.examples_path + "/example.beancount"
        converter = BeancountToLedgerConverter(debug=self.debug)
        beancount = converter.load(bc_example)
        self.assertIsInstance(beancount, Beancount)
        ledger_book = converter.convert_to_target()
        stats = ledger_book.get_stats()
        if self.debug:
            stats.show()
        example = NominaExample(
            name="example",
            owner="John Doe",
            url=None,
            example_path=Path(self.examples_path),
            do_log=True,
            expected_stats=Stats(
                accounts=63,
                transactions=966,
                start_date="2022-01-01",
                end_date="2024-10-04",
                currencies={
                    "EUR": 18,
                    "USD": 31,
                    "IRAUSD": 5,
                    "ITOT": 1,
                    "VEA": 1,
                    "VHT": 1,
                    "GLD": 1,
                    "VBMPX": 1,
                    "RGAGX": 1,
                    "VACHR": 3,
                },
            ),
        )
        wrong = example.check_stats(stats)
        self.assertEqual(0, wrong)

    def test_ledger2beancount(self):
        """
        test converting ledger format to Beancount
        """
        for name, example in self.examples.items():
            with self.subTest(f"Testing {name}"):
                l2b = LedgerToBeancountConverter()
                ledger_book = l2b.load(example.ledger_file)
                output_path = os.path.join("/tmp", f"{name}_l2b.beancount")
                beancount=l2b.convert_from_ledger(ledger_book)
                with open(output_path, "w") as bean_file:
                    output_text=l2b.to_text()
                    bean_file.write(output_text)
                if self.debug:
                    l2b.show_stats()
                # Verify the conversion
                self.assertIsNotNone(output_text)
                self.assertGreater(len(output_text), 0)

                # Parse the Beancount output to ensure it's valid
                beancount = Beancount()
                beancount.load_string(output_text)
                error_count = len(beancount.errors)
                if error_count > 0 and self.debug:
                    for i, error in enumerate(beancount.errors):
                        print(f"{i:3}:{error}")
                self.assertEqual(
                    error_count,
                    0,
                    f"Errors in Beancount conversion for {name}: {beancount.errors}",
                )

                # Compare statistics
                ledger_stats = l2b.source.get_stats()
                beancount_stats = beancount.get_stats()

                # Use the check_stats method from NominaExample
                example.check_stats(beancount_stats, ledger_stats)



"""
Created on 2024-10-06

@author: wf
"""

import os
import re
from dataclasses import dataclass
from typing import Dict, List, Optional

import chardet


@dataclass
class AccountingFileFormat:
    """
    A personal account file format
    """

    name: str
    acronym: str
    ext: str
    wikidata_id: str
    content_pattern: str
    encoding: str="utf-8"


class AccountingFileFormats:
    """
    Detector for various accounting file formats
    """

    def __init__(self):
        self.formats: List[AccountingFileFormat] = [
            AccountingFileFormat(
                name="Beancount",
                acronym="BEAN",
                ext=".beancount",
                wikidata_id="Q130456404",
                content_pattern=r'option "title"',
            ),
            AccountingFileFormat(
                name="GnuCash XML",
                acronym="GC-XML",
                ext=".gnucash",
                wikidata_id="Q130445392",
                content_pattern=r"<gnc-v2",
            ),
            AccountingFileFormat(
                name="GnuCash SQLite",
                acronym="GC-SQLITE",
                ext=".gnucash",
                wikidata_id="Q130445392",
                content_pattern=r"SQLite format 3",
            ),
            AccountingFileFormat(
                name="pyNomina Ledger Book YAML",
                acronym="LB-YAML",
                ext=".yaml",
                wikidata_id="Q281876",
                content_pattern=r"file_type:\s*NOMINA-LEDGER-BOOK-YAML|accounts:\s*\w+:",
            ),
            AccountingFileFormat(
                name="FinanzmanagerDeluxe",
                acronym="FMD",
                ext=".qif",
                wikidata_id="Q750657",
                content_pattern=r"!Option:MDY",
                encoding="iso-8859-1"
            ),
            AccountingFileFormat(
                name="Quicken Interchange Format",
                acronym="QIF",
                ext=".qif",
                wikidata_id="Q750657",
                content_pattern=r"!Account.*!Type:Cash",
            ),
            AccountingFileFormat(
                name="Subsembly JSON",
                acronym="BZV-JSON",
                ext=".json",
                wikidata_id="Q130443951",
                content_pattern=r'"AcctId":\s*"[^"]+".*"OwnrAcctCcy":\s*"[^"]+"',
            ),
            AccountingFileFormat(
                name="pyNomina Banking ZV YAML export",
                acronym="BZV-YAML",
                ext=".yaml",
                wikidata_id="Q281876",
                content_pattern=r"account_json_exports\s*:",
            ),
        ]
        self.format_by_acronym: Dict[str, AccountingFileFormat] = {
            fformat.acronym: fformat for fformat in self.formats
        }

    def get_by_acronym(self, acronym: str) -> "AccountingFileFormat":
        """
        get format by acronym
        """
        return self.format_by_acronym.get(acronym)

    def detect_format(self, file_path: str) -> Optional[AccountingFileFormat]:
        """
        Detect the accounting file format based on content and extension
        """
        _, file_extension = os.path.splitext(file_path)
        # Read the first chunk of the file
        with open(file_path, "rb") as file:
            raw_data = file.read(10000)  # Read first 10000 bytes

        # Detect encoding
        result = chardet.detect(raw_data)
        encoding = result["encoding"]

        # Decode the raw data
        try:
            content = raw_data.decode(encoding)
        except UnicodeDecodeError:
            # If the detected encoding fails, fall back to latin-1
            content = raw_data.decode("latin-1")

        for fformat in self.formats:
            if file_extension.lower() == fformat.ext.lower() and re.search(
                fformat.content_pattern, content, re.DOTALL | re.IGNORECASE
            ):
                return fformat

        return None

    def get_format_by_extension(self, file_path: str) -> Optional[AccountingFileFormat]:
        """
        Get the accounting file format based on file extension only
        """
        _, file_extension = os.path.splitext(file_path)

        for format in self.formats:
            if file_extension.lower() == format.ext.lower():
                return format

        return None



"""
Created on 2024-10-06

@author: wf
"""

from typing import TextIO, Optional
from lodstorage.persistent_log import Log
from nomina.file_formats import AccountingFileFormat, AccountingFileFormats
from nomina.ledger import Book

class AccountingFileConverter:
    """
    Abstract base class for accounting file converters.
    This class provides a structure for converting accounting data from one format to another.
    """

    def __init__(
        self,
        from_format_acronym: str,
        to_format_acronym: str,
        debug: bool = False,
    ):
        """
        Constructor that initializes the converter instance.

        Args:
            from_format_acronym (str): Acronym of the source format
            to_format_acronym (str): Acronym of the target format
            debug (bool): If True, enable debug output
        """
        formats = AccountingFileFormats()
        self.from_format: AccountingFileFormat = formats.get_by_acronym(from_format_acronym)
        self.to_format: AccountingFileFormat = formats.get_by_acronym(to_format_acronym)
        self.debug: bool = debug
        self.source: Optional[object] = None
        self.target: Optional[object] = None
        self.log: Log = Log()

    def load(self, input_stream: TextIO) -> object:
        """
        Load the content from the input stream.

        Args:
            input_stream (TextIO): The input stream

        Returns:
            object: The loaded content

        Raises:
            ValueError: If called on the abstract base class.
        """
        raise ValueError("load must be implemented in the subclass")

    def convert_to_target(self) -> object:
        """
        Convert the loaded source content to the target format.

        Returns:
            object: The converted content object (could be any personal accounting book type).

        Raises:
            ValueError: If called on the abstract base class.
        """
        raise ValueError("convert_to_target must be implemented in the subclass")

    def to_text(self) -> str:
        """
        Convert the target object into a text representation.

        Returns:
            str: The text representation of the target content.

        Raises:
            ValueError: If not implemented by a subclass.
        """
        raise ValueError("to_text must be implemented in the subclass")

    def show_stats(self) -> None:
        """
        Display statistics about the source and target objects.
        """
        source_stats = self.source.get_stats()
        if self.debug:
            source_stats.show()
        target_stats = self.target.get_stats()
        if self.debug:
            target_stats.show()

class BaseToLedgerConverter(AccountingFileConverter):
    """
    Base class for converters that convert from various formats to LedgerBook format.
    """

    def __init__(self, from_format_acronym: str, debug: bool = False):
        """
        Constructor for BaseToLedgerConverter.

        Args:
            from_format_acronym (str): Acronym of the source format
            debug (bool): If True, enable debug output
        """
        super().__init__(from_format_acronym, "LB-YAML", debug)

    def convert_to_ledger(self, input_path: str) -> Book:
        """
        Convert the input file to LedgerBook format.

        Args:
            input_path (str): Path to the input file

        Returns:
            Book: The converted LedgerBook object
        """
        if self.debug:
            print(f"Loading {self.from_format.name} from {input_path}")
        self.source = self.load(input_path)
        if self.debug:
            print(f"Converting {self.from_format.acronym} to {self.to_format.acronym}")
        self.target = self.convert_to_target()
        self.show_stats()
        return self.target

class BaseFromLedgerConverter(AccountingFileConverter):
    """
    Base class for converters that convert from LedgerBook format to various formats.
    """

    def __init__(self, to_format_acronym: str, debug: bool = False):
        """
        Constructor for BaseFromLedgerConverter.

        Args:
            to_format_acronym (str): Acronym of the target format
            debug (bool): If True, enable debug output
        """
        super().__init__("LB-YAML", to_format_acronym, debug)

    def convert_from_ledger(self, ledger_book: Book) -> object:
        """
        Convert from LedgerBook format to the target format object.

        Args:
            ledger_book (Book): The LedgerBook object to convert

        Returns:
            object: The converted object in the target format
        """
        self.set_source(ledger_book)
        if self.debug:
            print(f"Converting Ledger Book {ledger_book.name} to {self.to_format.acronym}")
        self.target = self.convert_to_target()
        self.show_stats()
        return self.target

    def to_text(self, target: object) -> str:
        """
        Convert the target object into a text representation.

        Args:
            target (object): The target object to convert to text

        Returns:
            str: The text representation of the target content.

        Raises:
            ValueError: If not implemented by a subclass.
        """
        raise ValueError("to_text must be implemented in the subclass")



"""
Created on 2024-10-06

@author: wf
"""

from copy import deepcopy

from nomina.qif_ledger import QifToLedgerConverter
from tests.basetest import Basetest
from tests.example_testcases import NominaExample


class Test_LedgerQif(Basetest):
    """
    test QIF handling
    """

    def setUp(self, debug=True, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.examples = NominaExample.get_examples(do_log=self.debug)

    def test_qif2ledger(self):
        """
        test the expenses example
        """
        name = "expenses"
        example = self.examples[name]
        qif_example = str(example.example_path) + "/expenses.qif"
        converter = QifToLedgerConverter()
        converter.load(qif_example)
        book = converter.convert_to_target()
        book.url = example.url
        book.owner = example.owner
        book.save_to_yaml_file(f"/tmp/{name}.yaml")
        stats = book.get_stats()
        if self.debug:
            stats.show()
        expected_stats = deepcopy(example.expected_stats)
        expected_stats.accounts += 2
        expected_stats.currencies["EUR"] += 5
        example.check_stats(stats, expected_stats=expected_stats)
        pass
